import json
from functools import lru_cache

from openai import OpenAI


def _simple_call(
    client: OpenAI,
    system_message: str | None,
    source_text: str,
    model: str = "gpt-3.5-turbo",
) -> str:
    messages = [{"role": "user", "content": source_text}]

    if system_message:
        messages.insert(0, {"role": "system", "content": system_message})

    chat_completion = client.chat.completions.create(
        messages=messages,
        model=model,
    )
    return chat_completion.choices[0].message.content


def _classification_call(
    client: OpenAI,
    system_message: str | None,
    source_text: str,
    categories: list[str],
    model: str = "gpt-3.5-turbo",
) -> str:
    messages = [{"role": "user", "content": source_text}]

    if system_message:
        messages.insert(0, {"role": "system", "content": system_message})

    tools = [
        {
            "type": "function",
            "function": {
                "name": "prediction",
                "description": "result of classification",
                "parameters": {
                    "type": "object",
                    "properties": {"result": {"type": "string", "enum": categories}},
                    "required": ["result"],
                },
            },
        }
    ]

    chat_completion = client.chat.completions.create(
        messages=messages,
        model=model,
        tools=tools,
        tool_choice={"type": "function", "function": {"name": "prediction"}},
    )
    args = chat_completion.choices[0].message.tool_calls[0].function.arguments
    return str(json.loads(args)["result"])


@lru_cache
def sentiment(src: str) -> str:
    client = OpenAI()
    return _classification_call(
        client,
        "Classify the sentiment expressed in the following text. The output should be one of 'positive', 'negative' or 'neutral'.",
        src,
        ["positive", "negative", "neutral"],
    )


@lru_cache
def summarize(src: str) -> str:
    client = OpenAI()
    return _simple_call(
        client,
        "Generate a short summary of a given text. Rely strictly on the provided text, without using any external knowledge.",
        src,
    )


@lru_cache
def ai(src: str) -> str:
    client = OpenAI()
    return _simple_call(client, None, src)



import duckdb
import pandas as pd

from sqllm.core import query
from sqllm.functions import sentiment


def test_sentiment():
    df = pd.DataFrame(
        {"text": ["this is awesome", "shit", "いいですね", None], "id": [1, 2, 3, 4]}
    )

    conn = duckdb.connect()
    conn.register("df", df)

    result = query(conn, "SELECT id, sentiment(text) as sentiment FROM df", [sentiment])

    assert result["sentiment"].values.tolist() == [
        "positive",
        "negative",
        "positive",
        None,
    ]


def test_ai():
    df = pd.DataFrame({"text": ["meow", "ワンワン", "bowwow", None], "id": [1, 2, 3, 4]})

    conn = duckdb.connect()
    conn.register("df", df)

    result = query(
        conn,
        'SELECT id, ai(\'If the next text is a dog sound, output "dog"; if it is a cat sound, output "cat". text: \' || text) as result FROM df',
    )

    assert result["result"].values.tolist() == ["cat", "dog", "dog", None]



import os
import tempfile
from typing import Any, Callable

import duckdb
import pandas as pd

from sqllm.functions import ai


def query(
    conn: Any,
    sql: str,
    functions: list[Callable[..., Any]] | None = None,
    null_on_error: bool = True,
) -> pd.DataFrame:
    """
    Executes an SQL query and returns the results in a DataFrame.
    The query can contain the `AI` function or user-defined functions (specified in the `functions` argument).

    :param conn: A database connection object (this is passed to pd.read_sql).
    :param sql: SQL query to be executed. Note that the query must conform to the SQL syntax of duckdb, not your DB.
    :param functions: List of user-defined functions. Functions must be type-annotated.
    :param null_on_error: Determines how query execution behaves when an error occurs inside a user-defined function;
                          if True (default), null is set to the result; if False, an exception is propagated.
    :return: Result of query.
    """
    functions = functions or []
    functions.append(ai)

    with tempfile.TemporaryDirectory() as dir:
        duckdb_conn = duckdb.connect(os.path.join(dir, "duckdb.db"))
        for function in functions:
            try:
                duckdb_conn.create_function(
                    function.__name__,
                    function,
                    null_handling="special",  # type: ignore
                    exception_handling="return_null" if null_on_error else "default",  # type: ignore
                )
            except Exception:
                pass
        tables = duckdb_conn.get_table_names(sql)

        for table in tables:
            df = pd.read_sql(f"SELECT * FROM {table}", conn)
            duckdb_conn.register("tmp_df", df)
            duckdb_conn.sql(f"CREATE TABLE {table} AS SELECT * FROM tmp_df")

        result = duckdb_conn.sql(sql)
        return result.df()


def query_df(
    df: pd.DataFrame,
    sql: str,
    functions: list[Callable[..., Any]] | None = None,
    null_on_error: bool = True,
) -> pd.DataFrame:
    """
    Executes an SQL query over the dataframe and returns the results in a DataFrame.
    The query can contain the `AI` function or user-defined functions (specified in the `functions` argument).

    :param df: A dataframe. In the query, you can refer to it as "df".
    :param sql: SQL query to be executed. Note that the query must conform to the SQL syntax of duckdb, not your DB.
    :param functions: List of user-defined functions. Functions must be type-annotated.
    :param null_on_error: Determines how query execution behaves when an error occurs inside a user-defined function;
                          if True (default), null is set to the result; if False, an exception is propagated.
    :return: Result of query.
    """
    conn = duckdb.connect()
    conn.register("df", df)
    return query(conn, sql, functions, null_on_error)



from datetime import date

import duckdb
import pandas as pd
from numpy.testing import assert_array_equal

from sqllm.core import query


def test_query_no_ai_function_simple():
    conn = duckdb.connect()

    df = pd.DataFrame(
        {
            "a": [1, 2, 3],
            "b": ["a", "b", "c"],
            "c": [date(2021, 1, 1), date(2022, 6, 15), date(2023, 12, 31)],
        }
    )

    conn.register("df", df)

    result = query(conn, "SELECT a, c FROM df WHERE a > 1")

    assert len(result) == 2
    assert list(result.columns) == ["a", "c"]


def test_query_no_ai_function_with_join():
    conn = duckdb.connect()

    user = pd.DataFrame({"user_id": [1, 2, 3, 4, 5], "age": [24, 58, 33, 40, 15]})
    orders = pd.DataFrame(
        {
            "order_id": [1, 2, 3, 4, 5, 6, 7],
            "user_id": [3, 4, 3, 2, 1, 1, 3],
            "item_id": [1, 2, 3, 1, 2, 3, 4],
            "price": [100, 200, 300, 100, 200, 350, 150],
        }
    )
    item = pd.DataFrame({"item_id": [1, 2, 3, 4], "name": ["a", "b", "c", "d"]})

    conn.register("user", user)
    conn.register("orders", orders)
    conn.register("item", item)

    result = query(
        conn,
        """
    SELECT
        user.user_id AS user_id,
        SUM(price) AS total_purchase,
        COUNT(DISTINCT item.item_id) AS item_variety,
    FROM
        orders
        LEFT JOIN user on orders.user_id = user.user_id
        LEFT JOIN item on orders.item_id = item.item_id
    GROUP BY
        user.user_id;
    """,
    )

    result = result.sort_values(by="user_id").reset_index(drop=True)

    assert_array_equal(result["user_id"], [1, 2, 3, 4])
    assert_array_equal(result["total_purchase"], [550, 100, 550, 200])
    assert_array_equal(result["item_variety"], [2, 1, 3, 1])


def test_query_ai_function():
    conn = duckdb.connect()

    df = pd.DataFrame(
        {
            "a": [1, 2, 3],
            "b": ["a", "b", "c"],
            "c": [date(2021, 1, 1), date(2022, 6, 15), date(2023, 12, 31)],
        }
    )

    conn.register("df", df)

    result = query(conn, "SELECT a, c, ai('hello!') FROM df WHERE a > 1")
    assert list(result.columns) == ["a", "c", "ai('hello!')"]
    assert "hello" in result[result.columns[-1]].iloc[0].lower()



from codecs import open
from os import path

from setuptools import find_packages, setup


def get_long_description():
    here = path.abspath(path.dirname(__file__))

    with open(path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = f.read()
    return long_description


def get_version():
    version_filepath = path.join(path.dirname(__file__), "sqllm", "version.py")
    with open(version_filepath) as f:
        for line in f:
            if line.startswith("__version__"):
                return line.strip().split()[-1][1:-1]


def requirements_from_file(filename):
    return open(filename).read().splitlines()


setup(
    name="sqllm",
    packages=find_packages(),
    version=get_version(),
    license="MIT",
    install_requires=requirements_from_file("requirements.txt"),
    author="nyanp",
    author_email="Noumi.Taiga@gmail.com",
    url="https://github.com/nyanp/sqllm",
    description="SQL with LLM",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    keywords="llm sql",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)
